package de.hendriklipka.config;

import java.util.Enumeration;

public interface Configuration
{
    /**
     *  Gets the RootBlock of the ConfigurationImpl object
     *
     *@return    The RootBlock value
     *
     * @post $return!=null
     */
    public Block getRootBlock();

    /**
     *  returns the block specified by the given name. Specifying sub-blocks (with
     *  "super.sub.block") is possible.
     *
     *@param  name  the block name
     *@return       the block
     *
     * @pre name!=null
     */
    public Block getBlockByName(String name);

    /**
     *  returns all blocks specified by the given name. Specifying sub-blocks (with
     *  "super.sub.block") is possible.
     *
     *@param  name  the name
     *@return       enumeration containing all block matching the name
     *
     * @pre name!=null
     */
    public Enumeration getBlocksByName(String name);

    /**
     *  reloads the configuration
     */
    public void reload() throws ConfigFileException;

    public void addSourceFile(String fileName);

    public boolean isAlreadyRead(String fileName);
}
