package de.hendriklipka.config.impl;

import org.apache.xerces.parsers.DOMParser;
import org.w3c.dom.*;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Observer;
import java.util.Vector;

import de.hendriklipka.config.Block;
import de.hendriklipka.config.Value;
import de.hendriklipka.config.ConfigFileException;

/**
 * @todo handle server types (or generic filter concept during read)
 * @todo proper exception and error handling
 * @todo handle observers
 */
public final class BlockImpl implements Block
{
    protected String _name="root";
    /**
     *  text value of this block
     */
    protected Value _value=null;

    protected BlockImpl _parent=null;

    /**
     *  child blocks for this block
     */
    protected Hashtable _childs=new Hashtable();

    protected String _sourceFile=null;
    protected String _baseDirectory=null;

    /**
     *  Gets the text value of the BlockImpl
     *
     *@return    The text value
     *
     * @post $result!=null
     */
    public Value getValue()
    {
        return _value;
    }

    /**
     *  Gets the specified child block
     *
     *@param  name  child block name
     *@return       the block (the first, if more than one exists)
     */
    public Block getChild(String name)
    {
        int pos=name.indexOf('.');
        if (-1==pos)
        {
            return getChildByName(name);
        }
        String cName=name.substring(0, pos);
        String sName=name.substring(pos+1);
        Block b=getChildByName(cName);
        if (null!=b)
            return b.getChild(sName);
        return null;
    }

    /**
     *  get all childs with the given name
     *
     *@param  name  block name
     *@return       enumeration with all block matching the given
     *
     * @post $result!=null
     */
    public Enumeration getChilds(String name)
    {
        int pos=name.indexOf('.');
        if (-1==pos)
        {
            Vector v=getChildsByName(name);
            if (null!=v)
                return v.elements();
            return null;
        }
        String cName=name.substring(0, pos);
        String sName=name.substring(pos+1);
        Block b=getChildByName(cName);
        if (null!=b)
            return b.getChilds(sName);
        return null;
    }

    /**
     *  adds an observer to this block
     *
     *@param  o  the observer to be added
     */
    public void addObserver(Observer o)
    {
    }

    /**
     *  removes an observer from this block
     *
     *@param  o  the observer to be removed
     */
    public void removeObserver(Observer o)
    {
    }

    public Enumeration getAllChilds()
    {
        Vector v=new Vector();
        Enumeration e=_childs.elements();
        while (e.hasMoreElements())
        {
            Vector vector=(Vector) e.nextElement();
            for (int i=0; i<vector.size(); i++)
            {
                Block block=(Block) vector.elementAt(i);
                v.addElement(block);
            }
        }
        return v.elements();
    }

    public String toString()
    {
        StringBuffer sb=new StringBuffer();
        sb.append(getFullName());
        sb.append("=");
        try
        {
            sb.append(_value.getValue());
            sb.append("\n");
            Enumeration e=getAllChilds();
            while (e.hasMoreElements())
            {
                Block block=(Block) e.nextElement();
                sb.append(block.toString());
            }
        }
        catch (ConfigFileException e1)
        {
            sb.append(e1.getMessage());
            sb.append("\n");
        }
        return sb.toString();
    }

    public String getFullName()
    {
        StringBuffer sb=new StringBuffer();
        if (null!=_parent)
        {
            sb.append(_parent.getFullName());
            sb.append(".");
        }
        sb.append(_name);

        return sb.toString();
    }

    public String getName()
    {
        return _name;
    }
    /**
     *  reads configuration from the given DOM node
     *
     *@param  node              the config node
     *@exception  org.xml.sax.SAXException  if an parser error happens
     *
     * @pre node!=null
     * @post value!=null
     */
    protected void readConfigurationFromXML(Node node, String baseDirectory, BlockImpl parent) throws ConfigFileException
    {
        _baseDirectory=baseDirectory;
        _parent=parent;
        Node attr;
        NamedNodeMap attrs=node.getAttributes();
        boolean isInclude=false;
        if (attrs!=null)
        {
            isInclude=readNodeAttributes(attrs, parent);
        }
        String text="";
        if (!isInclude)
        {
            NodeList nl=node.getChildNodes();
            text=readChildNodes(nl);
            text=text.trim();
        }
        _value=new ValueImpl(text, this);
    }

    private String readChildNodes(NodeList nl) throws ConfigFileException
    {
        StringBuffer textData=new StringBuffer();
        for (int i=0; i<nl.getLength(); i++)
        {
            Node subNode=nl.item(i);
            if (subNode.getNodeType()==Node.TEXT_NODE||subNode.getNodeType()==Node.CDATA_SECTION_NODE)
            {
                textData.append(subNode.getNodeValue());
            }
            else if (subNode.getNodeType()==Node.ELEMENT_NODE)
            {
                addChild(subNode, this);
            }
        }
        return textData.toString();
    }

    private boolean readNodeAttributes(NamedNodeMap attrs, BlockImpl parent) throws ConfigFileException
    {
        for (int i_attr=0; i_attr<attrs.getLength(); i_attr++)
        {
            Node attr=attrs.item(i_attr);
            if (attr.getNodeName().equals("include"))
            {
                String fName=attr.getNodeValue();
                _sourceFile=fName;
                readIncludeFile(fName, parent);
                return true;
            }
        }
        return false;
    }

    private void readIncludeFile(String fileName, BlockImpl parent) throws ConfigFileException
    {
        try
        {
            String fName=determineIncludeFilename(fileName);
            ConfigurationImpl.getInstance().addSourceFile(fName);
            String baseDir=new File(fName).getParentFile().getAbsolutePath();
            InputStream result=new FileInputStream(fName);
            DOMParser parser=new DOMParser();
            parser.parse(new InputSource(result));
            Document doc=parser.getDocument();
            Element el=doc.getDocumentElement();
            readConfigurationFromXML(el, baseDir, parent);
        }
        catch (SAXException e)
        {
            throw new ConfigFileException(e.getMessage());
        }
        catch (IOException e)
        {
            throw new ConfigFileException(e.getMessage());
        }
    }

    private String determineIncludeFilename(String fileName) throws ConfigFileException, IOException
    {
        File f=new File(fileName);
        if (!f.isAbsolute())
        {
            fileName=_baseDirectory+File.separator+fileName;
        }
        f=new File(fileName);
        if (!f.exists())
        {
            throw new ConfigFileException("cannot find included file "+fileName+" in "+_name);
        }
        fileName=f.getCanonicalPath();
        if (ConfigurationImpl.getInstance().isAlreadyRead(fileName))
        {
            throw new ConfigFileException("in "+_name+": source file "+fileName+" is already included");
        }
        return fileName;
    }

    /**
     *  Adds a child to the child list
     *
     *@param  name              name of the child
     *@param  node              DOM containing the child data
     *@exception  org.xml.sax.SAXException  if a parser error occurs
     *
     * @pre name!=null
     * @pre node!=null
     * @pre name.equals(node.getName())
     */
    private void addChild(Node node, BlockImpl parent) throws ConfigFileException
    {
        String name=node.getNodeName();
        Object o=_childs.get(name);
        if (null==o)
        {
            o=new Vector();
            _childs.put(name, o);
        }
        /** @assert v!=null, can't create new Vector object */
        Vector v=(Vector) o;
        BlockImpl b=new BlockImpl();
        b._name=name;
//        System.out.println("adding child "+name);
        b.readConfigurationFromXML(node, _baseDirectory, parent);
        v.addElement(b);
    }

    private Vector getChildsByName(String name)
    {
        Object o=_childs.get(name);
        if (null==o)
        {
            return null;
        }
        Vector v=(Vector) o;
        if (0==v.size())
        {
            return null;
        }
        return v;
    }

    private Block getChildByName(String name)
    {
        Vector v=getChildsByName(name);
        if (null!=v)
            return (Block) v.elementAt(0);
        return null;
    }

    Block getBlockViaParent(String name)
    {
        Block block=getChild(name);
        if (null!=block)
            return block;
        if (null!=_parent)
            return _parent.getBlockViaParent(name);
        return null;
    }
}
