/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along with this
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
hli@hendriklipka.de
*/

package de.hendriklipka.databus;

import java.util.*;
import java.io.*;

import org.w3c.dom.*;
import com.sun.xml.parser.Resolver;
import com.sun.xml.tree.XmlDocument;
import org.xml.sax.InputSource;
import org.apache.log4j.Category;
import org.apache.log4j.xml.DOMConfigurator;
import de.hendriklipka.config.impl.ConfigurationImpl;
import de.hendriklipka.config.Block;
import de.hendriklipka.config.Configuration;
import de.hendriklipka.config.ConfigFileException;

public class DataBus 
{
    static Category CAT = Category.getInstance(DataBus.class.getName());

    protected static Hashtable _descriptorsByName=new Hashtable();
    protected static Hashtable _descriptorsByClass=new Hashtable();

    protected String _databusName="unknown";

    protected Hashtable _allConsumers=new Hashtable();
    protected Hashtable _allProducers=new Hashtable();
    protected Hashtable _allComponents=new Hashtable();

    protected Vector _components=new Vector(5,5);
    protected Vector _genericProducers=new Vector(5,5);

    protected long _eventsInQueue=0;
    protected long _eventsInWork=0;
    protected Object _eventsInWorkSemaphore=new Object();
    protected boolean _doReConfig=false;
    protected Object _sendSyncObj=new Object();

    protected Hashtable _listenerTypes=new Hashtable();
    protected Hashtable _producerTypes=new Hashtable();

    public DataBus(Node configNode) throws Exception
    {
        NamedNodeMap nnl=configNode.getAttributes();
        Node attr=nnl.getNamedItem("name");
        _databusName=attr.getNodeValue();
        NodeList nl=configNode.getChildNodes();
        for (int i=0;i<nl.getLength();i++)
        {
            Node node=nl.item(i);
            if ((node.getNodeType()==node.ELEMENT_NODE)&&(node.getNodeName().equals("component")))
            {
                configureComponent(node);
            }
        }
        // add internal consumer for DataBusControlEvents
        addInternalComponents();

        // sort components according to their position
        Collections.sort(_components);
    }

    private void addInternalComponents()
    {
        DataBusControlComponent dcc=new DataBusControlComponent();
        ComponentDescriptor cd=new ComponentDescriptor();
        ComponentInstance ci=new ComponentInstance(); // create generic descriptor
        ci._id=dcc.toString();
        ci._component=dcc;
        ci._configuration=null;
        ci._descriptor=cd;
        _allComponents.put(dcc,ci);
        ci._position=Integer.MIN_VALUE; // start as first component
        _components.addElement(ci);
        _listenerTypes.put(DataBusControlEvent.NAME,DataBusControlEvent.NAME);
    }

    private void configureComponent(Node node) throws Exception
    {
        NamedNodeMap nnl;
        Node attr;
        String cname=null;
        nnl=node.getAttributes();
        String componentName=null;
        attr=nnl.getNamedItem("name");
        componentName=attr.getNodeValue();
        cname=getClassForComponent(componentName);
        CAT.debug("got class "+cname+" from component descriptor "+componentName);
        if (null==cname)
        {
            throw new Exception("<component> w/o component name!");
        }
        attr=nnl.getNamedItem("id");
        if (null==attr)
        {
            throw new Exception("<component> w/o id!");
        }
        CAT.info("trying to load class: "+cname);
        String id=attr.getNodeValue();
        Class c=Class.forName(cname);
        Object o=c.newInstance();
        DataBusComponent dbc=(DataBusComponent)o;
        ComponentDescriptor cd=getDescriptorForClass(cname,componentName);
        ComponentInstance ci=new ComponentInstance();
        ci._id=id;
        ci._component=dbc;
        ci._configuration=readComponentConfiguration(node.getNodeName(),node);
        ci._descriptor=cd;
        _allComponents.put(dbc,ci);
        attr=nnl.getNamedItem("position");
        if (null!=attr)
        {
            ci._position=Integer.parseInt(attr.getNodeValue());
        }
        _components.addElement(ci);
        CAT.info("added "+cname);
        if (dbc instanceof DataBusProducer)
        {
            configureAsProducer(dbc, cd, ci);
        }
        if (dbc instanceof DataBusConsumer)
        {
            configureAsConsumer(cd, dbc);
        }
    }

    private void configureAsConsumer(ComponentDescriptor cd, DataBusComponent dbc)
    {
        CAT.info("add as consument");
        String[] lta=cd._listenTo;
        if (null!=lta)
        {
            for (int j=0;j<lta.length;j++)
            {
                addConsumer((DataBusConsumer)dbc,lta[j]);
                _listenerTypes.put(lta[j],lta[j]);
            }
        }
    }

    private void configureAsProducer(DataBusComponent dbc, ComponentDescriptor cd, ComponentInstance ci)
    {
        CAT.info("add as producer: "+dbc+"/"+cd);
        _allProducers.put(dbc,cd);
        if (cd._isGenericProducer)
        {
            _genericProducers.addElement(ci);
        }
        String[] pta=cd._produces;
        if (null!=pta)
        {
            for (int j=0;j<pta.length;j++)
            {
                _producerTypes.put(pta[j],pta[j]);
            }
        }
    }

    /**
     * @todo add check for matching producers / consumers
     */
    public void start() throws Exception
    {
        // initialization
        CAT.debug("init");
        for (int i=0;i<_components.size();i++)
        {
            ComponentInstance ci=(ComponentInstance)_components.elementAt(i);
            CAT.info("position for "+ci._id+"="+ci._position);
            ci._component.init(this,ci._configuration);
        }
        // pre startup run
        CAT.debug("pre startup");
        for (int i=0;i<_components.size();i++)
        {
            ComponentInstance ci=(ComponentInstance)_components.elementAt(i);
            ci._component.preStartup();
        }
        // startup
        CAT.debug("startup");
        for (int i=0;i<_components.size();i++)
        {
            ComponentInstance ci=(ComponentInstance)_components.elementAt(i);
            ci._component.start();
        }
        //post startup run
        CAT.debug("post startup");
        for (int i=0;i<_components.size();i++)
        {
            ComponentInstance ci=(ComponentInstance)_components.elementAt(i);
            ci._component.postStartup();
        }
    }

    public String getProperty(String key, String def)
    {
        return _props.getProperty(key,def);
    }

    public String getProperty(String key)
    {
        return getProperty(key,null);
    }

    /**
        insert a new consumer into the bus, for a specific event
        @param dataType name of the event this consumer wants to receive events for (max. 15 chars)
    */
    public void addConsumer(DataBusConsumer dbc, String dataType)
    {
        if (dataType.length()>15)
            dataType=dataType.substring(0,15);
        Vector v=(Vector)_allConsumers.get(dataType);
        if (null==v)
        {
            v=new Vector(3,3);
            _allConsumers.put(dataType,v);
        }
        Consumer c=new Consumer(dbc,dbc.toString());
        v.addElement(c);
        new Thread(c).start();
    }

    /**
        removes a consumer
    */
    public synchronized void removeConsumer(DataBusConsumer dbc)
    {
        Enumeration e=_allConsumers.elements();
        while (e.hasMoreElements())
        {
            Vector v=(Vector)e.nextElement();
            for (int i=0;i<v.size();i++)
            {
                Consumer c=(Consumer)v.elementAt(i);
                if (dbc==c._dbc)
                {
                    v.removeElement(c);
                }
            }
        }
    }

    /**
        removes a component from the bus
    */
    public void componentFinished(DataBusComponent dbc)
    {
        _genericProducers.removeElement(dbc);
        _components.removeElement(dbc);
        _allComponents.remove(dbc);
        // look if there are generic producers left
        if (0==_genericProducers.size())
        {
            // no
            while(true)
            {
                // wait until there are no more events on the bus
                synchronized(_eventsInWorkSemaphore)
                {
                    if ((0==_eventsInQueue)&&(0==_eventsInWork))
                    {
                        break;
                    }
                    try
                    {
                        _eventsInWorkSemaphore.wait();
                    }
                    catch (InterruptedException ex)
                    {
                        ex.printStackTrace();
                    }
                }
            }
            Enumeration e=_allConsumers.elements();
            while (e.hasMoreElements())
            {
                Vector v=(Vector)e.nextElement();
                for (int i=0;i<v.size();i++)
                {
                    Consumer c=(Consumer)v.elementAt(i);
                    c._stop=true;
                    synchronized(c._notificationFlag)
                    {
                        c._notificationFlag.notifyAll();
                    }
                }
            }
            for (int i=0;i<_components.size();i++)
            {
                ComponentInstance ci=(ComponentInstance)_components.elementAt(i);
                DataBusComponent dbc2=(ci._component);
                try
                {
                    dbc2.shutdown();
                }
                catch (Exception ex)
                {
                    CAT.error("error during shutdown: ",ex);
                }
            }
        }
    }

    //ToDo
    protected void forceShutDown()
    {
    }

    /**
        distributes events on the bus
        @param data data to be send
        @param dataType type of the event
    */
    public void sendData(Object data,String dataType, DataBusProducer src)
    {
        synchronized(_sendSyncObj)
        {
            // during a reconfiguration we do not accept new data
            if (_doReConfig)
            {
                if (getDescriptor(src)._isGenericProducer)
                {
                    try
                    {
                        _sendSyncObj.wait();
                    }
                    catch (InterruptedException ex) {}
                }
            }
            if (dataType.length()>15)
                dataType=dataType.substring(0,15);
            Vector v=(Vector)_allConsumers.get(dataType);
            if (null==v)
            {
                return;
            }
            DataBusEvent dbe=new DataBusEvent();
            dbe._data=data;
            for (int i=0;i<v.size();i++)
            {
                ((Consumer)v.elementAt(i)).dataAvailable(dbe);
            }
        }
    }

    /**
        distributes data on the bus, but wait for an answer written to the event
        @param data data to be send
        @param dataType type of the event
        @param src source which waits for the answer
        @return the response object
    */
    public Object sendResponseData(Object data,String dataType, DataBusProducer src)
    {
        synchronized(_sendSyncObj)
        {
            // during a reconfiguration we do not accept new data
            if (_doReConfig)
            {
                if (getDescriptor(src)._isGenericProducer)
                {
                    try
                    {
                        _sendSyncObj.wait();
                    }
                    catch (InterruptedException ex) {}
                }
            }
            if (dataType.length()>15)
                dataType=dataType.substring(0,15);
            Vector v=(Vector)_allConsumers.get(dataType);
            if (null==v)
            {
                return null;
            }
            DataBusEvent dbe=new DataBusEvent();
            dbe._data=data;
            dbe._waitForResponse=true;
            for (int i=0;i<v.size();i++)
            {
                ((Consumer)v.elementAt(i)).dataAvailable(dbe);
            }
            // wait until waitForResponse has been reset
            // notification happans through dbe._notificationFlag
            synchronized(dbe._flag)
            {
                while(dbe._waitForResponse)
                {
                    try
                    {
                        dbe._flag.wait();
                    }
                    catch (InterruptedException ex)
                    {
                        ex.printStackTrace();
                    }
                }
            }
            return dbe._answer;
        }
    }

    protected void doReConfiguration(Node configNode)
    {
        /*
        CAT.debug("reconfiguring bus...");
        // lock bus
        synchronized(_sendSyncObj)
        {
            _doReConfig=true;
        }
        CAT.debug("waiting until bus has stopped...");
        while(true)
        {
            // wait until there are no more events
            synchronized(_eventsInWorkSemaphore)
            {
                if ((0==_eventsInQueue)&&(0==_eventsInWork))
                {
                    break;
                }
                try
                {
                    _eventsInWorkSemaphore.wait();
                }
                catch (InterruptedException ex)
                {
                    CAT.error("error during reconfiguration",ex);
                }
            }
        }

        try
        {
            NodeList nl=configNode.getChildNodes();
            for (int i=0;i<nl.getLength();i++)
            {
                Node node=nl.item(i);
                if ((node.getNodeType()==node.ELEMENT_NODE)&&(node.getNodeName().equals("component")))
                {
                    NamedNodeMap nnl=node.getAttributes();
                    Node attr=nnl.getNamedItem("id");
                    if (null==attr)
                    {
                        throw new Exception("<component> w/o id!");
                    }
                    // is this ID already there?
                    // yes -> reconfig
                    // no -> create
                    attr=nnl.getNamedItem("class");
                    if (null==attr)
                    {
                        throw new Exception("<component> w/o class!");
                    }
                    String cname=attr.getNodeValue();
                    String id=attr.getNodeValue();
                    Class c=Class.forName(cname);
                    Object o=c.newInstance();
                    DataBusComponent dbc=(DataBusComponent)o;
                    dbc.setID(id);
                    _components.addElement(dbc);
                    dbc.init(this,node);
                    if (dbc instanceof DataBusProducer)
                    {
                        if (((DataBusProducer)dbc).isGenericProducer())
                        {
                            _genericProducers.addElement(dbc);
                        }
                        String[] pta=((DataBusProducer)dbc).produces();
                        if (null!=pta)
                        {
                            for (int j=0;j<pta.length;j++)
                            {
                                _producerTypes.put(pta[j],pta[j]);
                            }
                        }
                    }
                    if (dbc instanceof DataBusConsumer)
                    {
                        String[] lta=((DataBusConsumer)dbc).listenTo();
                        if (null!=lta)
                        {
                            for (int j=0;j<lta.length;j++)
                            {
                                _listenerTypes.put(lta[j],lta[j]);
                            }
                        }
                    }
                }
            }
        }
        catch (Exception ex)
        {
            CAT.error("error during reconfiguration",ex);
        }

        CAT.debug("bus stopped");

        // unlock bus
        synchronized(_sendSyncObj)
        {
            _sendSyncObj.notifyAll();
        }
        CAT.debug("bus restarted");
        */
    }

    /**
        this class decouples a consumer from the bus, so that the bus is not blocked until the
        incoming events has been processed
        It uses an own thread, whichc stores the incoming events in an queue and sends them
        to the consumer. If the queue gets too long, the priority of this thread (the processing
        happans in its context) is increased (to have faster processing), until the queue gets an acceptable size.
    */
    protected class Consumer implements Runnable
    {
        String _componentName="";
        Vector _allEvents=new Vector(5,5);
        Object _notificationFlag=new Object();
        DataBusConsumer _dbc=null;
        boolean _stop=false;

        Consumer (DataBusConsumer dbc, String name)
        {
            _dbc=dbc;
            _componentName=name;
        }

        public void run()
        {
            while (!_stop)
            {
                DataBusEvent dbe=getNextEvent();
                // if event is available, handle it
                if (null!=dbe)
                {
                    handleEvent(dbe);
                }
                // if no event is available, go to sleep
                else
                {
                    synchronized(_notificationFlag)
                    {
                        try
                        {
                            _notificationFlag.wait();
                        }
                        catch (InterruptedException ex)
                        {
                            CAT.error("got interrupted during waiting for answer",ex);
                        }
                    }
                }
            }
        }

        private void handleEvent(DataBusEvent dbe)
        {
            try
            {
                synchronized (_eventsInWorkSemaphore)
                {
                    _eventsInWork++;
                }
                // send event to consumer
                _dbc.dataAvailable(dbe);
                synchronized (_eventsInWorkSemaphore)
                {
                    _eventsInWork--;
                    _eventsInWorkSemaphore.notifyAll();
                }
                if (_allEvents.size()<5)
                    Thread.currentThread().yield();
                rePrioritize();
            }
            catch (Throwable ex)
            {
                CAT.error("error during data delivery: ",ex);
            }
            if (dbe._waitForResponse)
            {
                synchronized(dbe._flag)
                {
                    dbe._waitForResponse=false;
                    dbe._flag.notifyAll();
                }
            }
        }

        private DataBusEvent getNextEvent()
        {
            DataBusEvent dbe=null;
            // try to get the first event from the queue
            synchronized(_notificationFlag)
            {
                if (0!=_allEvents.size())
                {
                    synchronized(_allEvents)
                    {
                        dbe=(DataBusEvent)_allEvents.elementAt(0);
                        _allEvents.removeElementAt(0);
                        _eventsInQueue--;
                    }
                }
            }
            return dbe;
        }

        private void rePrioritize()
        {
            //ToDo: increase priority more than once, if neccessary
            try
            {
                boolean highPri=false;
                // if the queue is too long, increase priority
                if ((_allEvents.size()>100)&&(!highPri))
                {
                    highPri=true;
                    Thread t=Thread.currentThread();
                    int i=t.getPriority();
                    t.setPriority(i+1);
                }
                // if we have increased priority and the queue is short enough,
                // reduce priority
                if ((highPri)&&(_allEvents.size()<10))
                {
                    highPri=false;
                    Thread t=Thread.currentThread();
                    int i=t.getPriority();
                    t.setPriority(i-1);
                }
            }
            catch (Exception ex)
            {
                ex.printStackTrace();
            }
        }

        /**
         * add new event to handle. wakes up thew handler thread
         * @param dbe the event
         */ 
        void dataAvailable(DataBusEvent dbe)
        {
            synchronized(_notificationFlag)
            {
                _allEvents.addElement(dbe);
                _eventsInQueue++;
                _notificationFlag.notifyAll();
            }
        }
    }

    protected ComponentDescriptor getDescriptorForClass(String classname, String componentName)
    {
        CAT.debug("get descriptor for class "+classname);
        Object o=_descriptorsByClass.get(classname);
        if (null==o)
        {
            if (null!=componentName)
            {
                ComponentDescriptor cd=lookupComponent(componentName);
                if (null!=cd)
                    return cd;
            }
            CAT.warn("no descriptor found for class "+classname);
        }
        return (ComponentDescriptor)o;
    }

    protected ComponentDescriptor getDescriptorForName(String name)
    {
        CAT.debug("get descriptor for name "+name);
        return (ComponentDescriptor)_descriptorsByName.get(name);
    }

    protected static ComponentDescriptor registerComponent(Node node)
    {
        ComponentDescriptor cd=new ComponentDescriptor(node);
        CAT.info("registering component <"+cd._componentName+">, class <"+cd._componentClass+">");
        _descriptorsByClass.put(cd._componentClass,cd);
        _descriptorsByName.put(cd._componentName,cd);
        return cd;
    }

    protected ComponentDescriptor getDescriptor(DataBusProducer comp)
    {
        return (ComponentDescriptor)_allProducers.get(comp);
    }

    protected String getClassForComponent(String name)
    {
        CAT.debug("get class for descriptor "+name);
        ComponentDescriptor cd=lookupComponent(name);
        if (null!=cd)
            return cd._componentClass;
        return null;
    }
    
    protected ComponentDescriptor lookupComponent(String name)
    {
        ComponentDescriptor cd=(ComponentDescriptor)_descriptorsByName.get(name);
        if (null!=cd)
        {
            return cd;
        }
        CAT.debug("looking for component descriptor file "+name+".xml");
        InputStream is=this.getClass().getClassLoader().getResourceAsStream(name+".xml");
        if (null!=is)
        {
            try
            {
                
                InputSource input = new InputSource(is);
                XmlDocument doc = XmlDocument.createXmlDocument(input, false);
                Element el=doc.getDocumentElement();
                return registerComponent(el);
            }
            catch (Exception e)
            {
                e.printStackTrace();
            }
        }
        return null;
    }

    protected ComponentInstance getInstanceForComponent(DataBusComponent dbc)
    {
        return (ComponentInstance)_allComponents.get(dbc);
    }

    protected Block readComponentConfiguration(String elementName, Node configNode) throws ConfigFileException
    {
        Configuration config=ConfigurationImpl.createConfigurationFromNode(configNode);
        Block block=config.getRootBlock();
        
        return block;
    }

//-----------------------------------------------------------------------------
// static attributes, classes and methods

    /**
        waiting for a modified config file
    */
    protected static class ConfigChangeThread extends Thread
    {
        protected Object _waitObject=new Object();
        protected String _fileName=null;
        protected long _fileDate=0;
        protected long _sleepTime=15*1000;

        public ConfigChangeThread(String fileName, int sleepTime)
        {
            _fileName=fileName;
            _sleepTime=sleepTime;
            setDaemon(true);
            File f=new File(_fileName);
            _fileDate=f.lastModified();
        }

        public void run()
        {
            CAT.info("config file watcher started");
            while (true)
            {
                synchronized(_waitObject)
                {
                    try
                    {
                        _waitObject.wait(_sleepTime);
                    }
                    catch (InterruptedException ex) {}
                }
                File f=new File(_fileName);
                long d=f.lastModified();
                if (d>_fileDate)
                {
                    // execute reconfiguration
                    CAT.info("detected new config file");
                    _fileDate=d;
                    doReConfiguration(_fileName);
                }
            }
        }
    }

    // reconfig all databusses
    protected static void doReConfiguration(String fileName)
    {
        Vector newBusses=new Vector(5,5);
        try
        {
            // open config file
            InputSource input = Resolver.createInputSource(new File(fileName));
            XmlDocument doc = XmlDocument.createXmlDocument(input, false);
            Element el=doc.getDocumentElement();

            /*
            // we don't reconfig log4j because then the logfile is closed and re-opened :(
            NodeList cnl=el.getElementsByTagName("log4j");
            Node configNode=cnl.item(0);
            DOMConfigurator.configure((Element)configNode);
            */

            // read all busses
            NodeList nl=el.getElementsByTagName("databus");
            for (int i=0;i<nl.getLength();i++)
            {
                Node node=nl.item(i);
                NamedNodeMap nnl=node.getAttributes();
                Node attr=nnl.getNamedItem("name");
                String name=attr.getNodeValue();
                // do we already have this bus ?
                DataBus db=null;
                for (int j=0;j<_busses.size();j++)
                {
                    DataBus db1=(DataBus)_busses.elementAt(j);
                    if (name.equals(db1._databusName))
                    {
                        db=db1;
                        break;
                    }
                }
                if (null!=db)
                {
                    // yes
                    db.doReConfiguration(node);
                }
                else
                {
                    // no -> create new bus
                    db=new DataBus(node);
                    _busses.addElement(db);
                    db.start();
                }
                newBusses.addElement(name);
            }
            // look for existing busses which are not in the file anymore
            for (int i=0;i<_busses.size();i++)
            {
                String name1=((DataBus)_busses.elementAt(i))._databusName;
                boolean found=false;
                for (int j=0;j<newBusses.size();j++)
                {
                    String name2=(String)newBusses.elementAt(i);
                    if (name2.equals(name1))
                    {
                        found=true;
                        break;
                    }
                }
                if (!found)
                {
                    // stop this bus
                    DataBus db=(DataBus)_busses.elementAt(i);
                    db.forceShutDown();
                    _busses.removeElement(db);
                }
            }
        }
        catch (Exception ex)
        {
            CAT.error("error during reconfiguration :",ex);
        }
    }

    protected static Properties _props=new Properties();
    protected static Vector _busses=new Vector(5,5);

    public static void main(String[] args)
    {
        try
        {
            String fName="config.xml";
            fName=readCommandLineArguments(args, fName);

            InputSource input = Resolver.createInputSource(new File(fName));
            XmlDocument doc = XmlDocument.createXmlDocument(input, false);
            Element el=doc.getDocumentElement();

            NodeList cnl=el.getElementsByTagName("log4j:configuration");
            Node configNode=cnl.item(0);
            DOMConfigurator.configure((Element)configNode);

            registerBusses(el);
            
            ConfigChangeThread cct=new ConfigChangeThread(fName,10000);
            cct.start();
        }
        catch (Throwable ex)
        {
            CAT.error("error during init:",ex);
            System.exit(1);
        }
    }

    private static void registerBusses(Element el) throws Exception
    {
        // after that, create all busses
        NodeList nl=el.getElementsByTagName("databus");
        for (int i=0;i<nl.getLength();i++)
        {
            Node node=nl.item(i);
            DataBus db=new DataBus(node);
            _busses.addElement(db);
        }
        for (int i=0;i<_busses.size();i++)
        {
            DataBus db=(DataBus)(_busses.elementAt(i));
            db.start();
        }
    }

    private static String readCommandLineArguments(String[] args, String fName)
    {
        if ((null!=args)&&(0!=args.length))
        {
            fName=args[0];
            if (1<args.length)
            {
                for (int i=1;i<args.length;i++)
                {
                    try
                    {
                        String s=args[i];
                        StringTokenizer st=new StringTokenizer(s,"=");
                        String name=(String)st.nextElement();
                        String val=(String)st.nextElement();
                        CAT.info("Property: "+name+"="+val);
                        _props.put(name,val);
                    }
                    catch (Exception ex)
                    {
                        System.err.println("error while parsing command line properties: no '=' found in parameter, use \"");
                        System.exit(1);
                    }
                }
            }
        }
        return fName;
    }

    public static String getTextData(Node node)
    {
        NodeList subs=node.getChildNodes();
        String s="";
        for (int j=0;j<subs.getLength();j++)
        {
            Node attr=subs.item(j);
            if ((attr.getNodeType()!=attr.TEXT_NODE)&&(attr.CDATA_SECTION_NODE!=attr.getNodeType()))
                continue;
            Text val=(Text)attr;
            s+=val.getData().trim();
        }
        if (0==s.length())
            return null;
        return s;
    }
}
