/*
Copyright (C) 1999/2000 Hendrik Lipka.
All rights reserved.

This program is free software; you can redistribute it and/or modify it under the terms
of the GNU General Public License as published by the Free Software Foundation;
either version 2 of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or
FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
for more details.

You should have received a copy of the GNU General Public License along with this
program; if not, write to the Free Software Foundation, Inc., 675 Mass Ave,
Cambridge, MA 02139, USA or see http://www.gnu.org

If you encounter problems or have a suggestion, please mail to:
hli@hendriklipka.de
*/

package de.hendriklipka.databus;

import de.hendriklipka.config.Block;

/**
    public interface for all members of a databus
*/
public interface DataBusComponent
{
    /**
        initialize this component
        if a component is a listener, it should register in this function
        @param bus the databus of which this component is a member
        @param configuration the XML node with the configuration data for this component
    */
    public void init(DataBus bus, Block configuration) throws Exception;

    /**
        starts this component
        this function is called when all databus member are initialized, but before they are started
    */
    public void preStartup() throws Exception;

    /**
        starts this component
        this function is called when all databus member are initialized
        threads should be started in this function
    */
    public void start() throws Exception;

    /**
        starts this component
        this function is called when all databus member are started
    */
    public void postStartup() throws Exception;

    /**
        is called if the databus goes down
        after the call to this function no new events should be produced, except
        if they are based on events received from other components
    */
    public void shutdown() throws Exception;

    /**
        is called, if this component is removed from the databus
        this happens, if this component does not exist anymore after a databus
        reconfiguration or during shutdown, if all components have received shutdown() and
        there are no more events on the bus
    */
    public void deInitialize() throws Exception;

    /**
        called if the configuration file has been read again
        during the calls to this function the event distribution is stopped
        and there are no more events on the bus
    */
    public void reconfig(Block configuration) throws Exception;
}
